/*
 * fuzzuf
 * Copyright (C) 2021 Ricerca Security
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 */
/**
 * @file random.hpp
 * @author Ricerca Security <fuzzuf-dev@ricsec.co.jp>
 */
#ifndef FUZZUF_INCLUDE_ALGORITHM_LIBFUZZER_RANDOM_HPP
#define FUZZUF_INCLUDE_ALGORITHM_LIBFUZZER_RANDOM_HPP
#include <cassert>
#include <cctype>
#include <type_traits>
namespace fuzzuf::algorithm::libfuzzer {
namespace detail {

/**
 * Return random value in [min, max)
 *
 * Corresponding code of original libFuzzer implementation
 * https://github.com/llvm/llvm-project/blob/llvmorg-12.0.1/compiler-rt/lib/fuzzer/FuzzerRandom.h#L21
 *
 * @tparam V Integral type
 * @tparam RNG A type that satisfies standard random number generator concept.
 * @param rng Random number generator
 * @param min Minimum value(inclusive)
 * @param max Maximum value(exclusive)
 */
template <typename V, typename RNG>
auto random_value(RNG &rng, V min, V max)
    -> std::enable_if_t<std::is_integral_v<V>, V> {
  assert(min < max);
  if (min == max)
    return 0;
  return min + (rng() % (max - min));
}

} // namespace detail

/**
 * libFuzzer compatible random number distribution for integer with max
 * In spite of it is bad practice to use lower bits of values generated by linear congruential generators, libFuzzer is using such distribution.
 *
 * Corresponding code of original libFuzzer implementation
 * https://github.com/llvm/llvm-project/blob/llvmorg-12.0.1/compiler-rt/lib/fuzzer/FuzzerRandom.h#L28
 *
 * @tparam V Integral type
 * @tparam RNG A type that satisfies standard random number generator concept
 * @param rng Random number generator
 * @param max Maximum value (exclusive)
 */
template <typename V, typename RNG>
auto random_value(RNG &rng, V max)
    -> std::enable_if_t<std::is_integral_v<V>, V> {
  return max ? (rng() % max) : V(0);
}

/**
 * libFuzzer compatible random number distribution for boolean
 *
 * Corresponding code of original libFuzzer implementation
 * https://github.com/llvm/llvm-project/blob/llvmorg-12.0.1/compiler-rt/lib/fuzzer/FuzzerRandom.h#L22
 *
 * @tparam V Integral
 * @tparam RNG A type that satisfies standard random number generator concept
 * @param rng Random number generator
 */
template <typename V, typename RNG>
auto random_value(RNG &rng)
    -> std::enable_if_t<std::is_integral_v<V> && std::is_same_v<bool, V>, V> {
  return detail::random_value(rng, 0, 2);
}

/**
 * libFuzzer compatible random number distribution for integer without max
 *
 * @tparam V Integral
 * @tparam RNG A type that satisfies standard random number generator concept
 * @param rng Random number generator
 */
template <typename V, typename RNG>
auto random_value(RNG &rng)
    -> std::enable_if_t<std::is_integral_v<V> && !std::is_same_v<bool, V>, V> {
  return detail::random_value(rng, std::numeric_limits<V>::min(),
                              std::numeric_limits<V>::max());
}

} // namespace fuzzuf::algorithm::libfuzzer
#endif
